from synapse_sdk.clients.base import BaseClient
from synapse_sdk.clients.utils import get_default_url_conversion


class MLClientMixin(BaseClient):
    def list_models(self, params=None):
        path = 'models/'
        return self._list(path, params=params)

    def get_model(self, pk, params=None, url_conversion=None):
        path = f'models/{pk}/'
        url_conversion = get_default_url_conversion(url_conversion, files_fields=['file'], is_list=False)
        return self._get(path, params=params, url_conversion=url_conversion)

    def create_model(self, data):
        path = 'models/'
        file = data.pop('file')
        data['chunked_upload'] = self.create_chunked_upload(file)['id']
        return self._post(path, data=data)

    def list_ground_truth_events(self, params=None, url_conversion=None, list_all=False):
        path = 'ground_truth_events/'
        url_conversion = get_default_url_conversion(url_conversion, files_fields=['files'])
        return self._list(path, params=params, url_conversion=url_conversion, list_all=list_all)

    def get_ground_truth_version(self, pk):
        path = f'ground_truth_dataset_versions/{pk}/'
        return self._get(path)



from synapse_sdk.clients.base import BaseClient
from synapse_sdk.clients.exceptions import ClientError


class ServeClientMixin(BaseClient):
    def list_serve_applications(self, params=None):
        path = 'api/serve/applications/'
        response = self._get(path, params=params)
        for key, item in response['applications'].items():
            response['applications'][key]['deployments'] = list(item['deployments'].values())
        return list(response['applications'].values())

    def get_serve_application(self, pk, params=None):
        path = 'api/serve/applications/'
        response = self._get(path, params=params)
        try:
            response['applications'][pk]['deployments'] = list(response['applications'][pk]['deployments'].values())
            return response['applications'][pk]
        except KeyError:
            raise ClientError(404, 'Serve Application Not Found')



class ClientError(Exception):
    status = None
    reason = None

    def __init__(self, status, reason, *args):
        self.status = status
        self.reason = reason
        super().__init__(status, reason, *args)



import click

from .utils import CONFIG_DIR, DEFAULT_FILE, ensure_config_dir, get_default_alias


@click.command()
def list():
    """List all aliases."""
    ensure_config_dir()
    aliases = [f.stem for f in CONFIG_DIR.glob('*') if f.is_file() and f.stem != DEFAULT_FILE.stem]
    default_alias = get_default_alias()
    if not aliases:
        click.echo('No aliases found.')
        return
    for alias_name in aliases:
        if alias_name == default_alias:
            click.echo(f'{alias_name} (default)')
        else:
            click.echo(alias_name)



import os
from functools import cached_property
from typing import Any, Dict

from synapse_sdk.clients.backend import BackendClient
from synapse_sdk.loggers import BackendLogger, ConsoleLogger
from synapse_sdk.plugins.utils import read_plugin_config
from synapse_sdk.shared.enums import Context
from synapse_sdk.utils.storage import get_storage
from synapse_sdk.utils.string import hash_text


class PluginRelease:
    config: Dict[str, Any]
    envs = None

    def __init__(self, config=None, plugin_path=None, envs=None):
        if config:
            self.config = config
        else:
            self.config = read_plugin_config(plugin_path=plugin_path)
        self.envs = envs

    @cached_property
    def plugin(self):
        return self.config['code']

    @cached_property
    def version(self):
        return self.config['version']

    @cached_property
    def code(self):
        return f'{self.plugin}@{self.version}'

    @cached_property
    def category(self):
        return self.config['category']

    @cached_property
    def name(self):
        return self.config['name']

    @cached_property
    def package_manager(self):
        return self.config.get('package_manager', 'pip')

    @cached_property
    def package_manager_options(self):
        options = {'pip': {}, 'uv': {'uv_pip_install_options': []}}
        return options[self.package_manager]

    @cached_property
    def checksum(self):
        return hash_text(self.code)

    @cached_property
    def actions(self):
        return list(self.config['actions'].keys())

    def setup_runtime_env(self):
        import ray
        from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy
        from ray.util.state import list_nodes

        @ray.remote
        def warm_up():
            pass

        nodes = list_nodes(address=self.envs['RAY_DASHBOARD_URL'])
        node_ids = [n['node_id'] for n in nodes]
        for node_id in node_ids:
            strategy = NodeAffinitySchedulingStrategy(node_id=node_id, soft=False)

            warm_up.options(
                runtime_env={
                    self.package_manager: {
                        'packages': ['-r ${RAY_RUNTIME_ENV_CREATE_WORKING_DIR}/requirements.txt']
                        ** self.package_manager_options
                    },
                    'working_dir': self.get_url(self.envs['SYNAPSE_PLUGIN_STORAGE']),
                },
                scheduling_strategy=strategy,
            ).remote()

    def get_action_config(self, action):
        return self.config['actions'][action]

    def get_url(self, storage_url):
        storage = get_storage(storage_url)
        return storage.get_url(f'{self.checksum}.zip')

    def get_serve_url(self, serve_address, path):
        return os.path.join(serve_address, self.checksum, path)


class Run:
    """Run class for manage plugin run istance.

    Attrs:
        job_id: plugin run job id
        context: plugin run context
        client: backend client for communicate with backend
        logger: logger for log plugin run events
    """

    logger = None
    job_id = None
    context = None
    client = None

    def __init__(self, job_id, context):
        self.job_id = job_id
        self.context = context
        self.client = BackendClient(
            self.context['envs']['SYNAPSE_PLUGIN_RUN_HOST'],
            token=self.context['envs'].get('SYNAPSE_PLUGIN_RUN_USER_TOKEN'),
            tenant=self.context['envs'].get('SYNAPSE_PLUGIN_RUN_TENANT'),
        )
        self.set_logger()

    def set_logger(self):
        kwargs = {'progress_categories': self.context['progress_categories']}

        if self.job_id:
            self.logger = BackendLogger(self.client, self.job_id, **kwargs)
        else:
            self.logger = ConsoleLogger(**kwargs)

    def set_progress(self, current, total, category=''):
        """Set progress for plugin run.

        Args:
            current: current progress value
            total: total progress value
            category: progress category
        """
        self.logger.set_progress(current, total, category)

    def log(self, event, data, file=None):
        self.logger.log(event, data, file=file)

    def log_message(self, message, context=Context.INFO.value):
        self.logger.log('message', {'context': context, 'content': message})

    def end_log(self):
        self.log_message('Plugin run is complete.')



import json
from datetime import datetime
from enum import Enum
from typing import Annotated, Dict, List

from pydantic import AfterValidator, BaseModel, field_validator
from pydantic_core import PydanticCustomError

from synapse_sdk.clients.exceptions import ClientError
from synapse_sdk.clients.utils import get_batched_list
from synapse_sdk.clients.validators.collections import FileSpecificationValidator
from synapse_sdk.i18n import gettext as _
from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod
from synapse_sdk.plugins.models import Run
from synapse_sdk.shared.enums import Context
from synapse_sdk.utils.pydantic.validators import non_blank
from synapse_sdk.utils.storage import get_pathlib


class UploadStatus(str, Enum):
    SUCCESS = 'success'
    FAILED = 'failed'


class UploadRun(Run):
    class DataFileLog(BaseModel):
        """Data file log model."""

        data_file_info: str | None
        status: UploadStatus
        created: str

    class DataUnitLog(BaseModel):
        """Data unit log model."""

        data_unit_id: int | None
        status: UploadStatus
        created: str

    class TaskLog(BaseModel):
        """Task log model."""

        task_id: int | None
        status: UploadStatus
        created: str

    def log_data_file(self, data_file_info: dict, status: UploadStatus):
        """Upload data_file log.

        Args:
            data_file_info (dict): The json info of the data file.
            checksum (str): The checksum of the data file.
            status (DataUnitStatus): The status of the data unit.
        """
        now = datetime.now().isoformat()
        self.log(
            'upload_data_file',
            self.DataFileLog(data_file_info=json.dumps(data_file_info), status=status.value, created=now).model_dump(),
        )

    def log_data_unit(self, data_unit_id: int, status: UploadStatus):
        """Upload data_unit log.

        Args:
            data_unit_id (int): The ID of the data unit.
            status (DataUnitStatus): The status of the data unit.
        """
        now = datetime.now().isoformat()
        self.log(
            'upload_data_unit',
            self.DataUnitLog(data_unit_id=data_unit_id, status=status.value, created=now).model_dump(),
        )

    def log_task(self, task_id: int, status: UploadStatus):
        """Upload task log.

        Args:
            task_id (int): The ID of the task.
            status (UploadStatus): The status of the task.
        """
        now = datetime.now().isoformat()
        self.log('upload_task', self.TaskLog(task_id=task_id, status=status.value, created=now).model_dump())


class UploadParams(BaseModel):
    """Upload action parameters.

    Args:
        name (str): The name of the action.
        description (str | None): The description of the action.
        checkpoint (int | None): The checkpoint of the action.
        path (str): The path of the action.
        storage (int): The storage of the action.
        collection (int): The collection of the action.
        project (int | None): The project of the action.
        is_generate_tasks (bool): The flag to generate tasks.
        is_generate_ground_truths (bool): The flag to generate ground truths
    """

    name: Annotated[str, AfterValidator(non_blank)]
    description: str | None
    path: str
    storage: int
    collection: int
    project: int | None
    is_generate_tasks: bool = False
    is_generate_ground_truths: bool = False

    @field_validator('storage', mode='before')
    @classmethod
    def check_storage_exists(cls, value: str, info) -> str:
        """Validate synapse-backend storage exists.

        TODO: Need to define validation method naming convention.
        TODO: Need to make validation method reusable.
        """
        action = info.context['action']
        client = action.client
        try:
            client.get_storage(value)
        except ClientError:
            raise PydanticCustomError('client_error', _('Error occurred while checking storage exists.'))
        return value

    @field_validator('collection', mode='before')
    @classmethod
    def check_collection_exists(cls, value: str, info) -> str:
        """Validate synapse-backend collection exists."""
        action = info.context['action']
        client = action.client
        try:
            client.get_dataset(value)
        except ClientError:
            raise PydanticCustomError('client_error', _('Error occurred while checking collection exists.'))
        return value

    @field_validator('project', mode='before')
    @classmethod
    def check_project_exists(cls, value: str, info) -> str:
        """Validate synapse-backend project exists."""
        if not value:
            return value

        action = info.context['action']
        client = action.client
        try:
            client.get_project(value)
        except ClientError:
            raise PydanticCustomError('client_error', _('Error occurred while checking project exists.'))
        return value


@register_action
class UploadAction(Action):
    """Upload action class.

    Attrs:
        name (str): The name of the action.
        category (PluginCategory): The category of the action.
        method (RunMethod): The method to run of the action.

    Progress Categories:
        analyze_collection: The progress category for the analyze collection process.
        data_file_upload: The progress category for the upload process.
        generate_data_units: The progress category for the generate data units process.
        generate_tasks: The progress category for the generate tasks process.
        generate_ground_truths: The progress category for the generate ground truths process.
    """

    name = 'upload'
    category = PluginCategory.UPLOAD
    method = RunMethod.JOB
    run_class = UploadRun
    progress_categories = {
        'analyze_collection': {
            'proportion': 5,
        },
        'upload_data_files': {
            'proportion': 35,
        },
        'generate_data_units': {
            'proportion': 20,
        },
        'generate_tasks': {
            'proportion': 20,
        },
        'generate_ground_truths': {
            'proportion': 20,
        },
    }

    def get_uploader(self, path):
        """Get uploader from entrypoint."""
        return self.entrypoint(self.run, path)

    def start(self) -> Dict:
        """Start upload process.

        Returns:
            Dict: The result of the upload process.
        """
        # Setup path object with path and storage.
        storage = self.client.get_storage(self.params['storage'])
        pathlib_cwd = get_pathlib(storage, self.params['path'])

        # Initialize uploader.
        uploader = self.get_uploader(pathlib_cwd)

        # Analyze Collection file specifications to determine the data structure for upload.
        file_specification_template = self._analyze_collection()

        # Setup result dict.
        result = {}

        # Organize data according to Collection file specification structure.
        organized_files = uploader.handle_upload_files()
        if not self._validate_organized_files(file_specification_template, organized_files):
            self.run.log_message('Validate organized files failed.')
            self.run.end_log()
            return result

        # Upload files to synapse-backend.
        organized_files_count = len(organized_files)
        if not organized_files_count:
            self.run.log_message('Files not found on the path.', context=Context.WARNING.value)
            self.run.end_log()
            return result
        uploaded_files = self._upload_files(organized_files, organized_files_count)
        result['uploaded_files_count'] = len(uploaded_files)

        # Generate data units for the uploaded data.
        upload_result_count = len(uploaded_files)
        if not upload_result_count:
            self.run.log_message('No files were uploaded.', context=Context.WARNING.value)
            self.run.end_log()
            return result
        generated_data_units = self._generate_data_units(uploaded_files, upload_result_count)
        result['generated_data_units_count'] = len(generated_data_units)

        # Setup task with uploaded synapse-backend data units.
        if not len(generated_data_units):
            self.run.log_message('No data units were generated.', context=Context.WARNING.value)
            self.run.end_log()
            return result

        if self.config['options']['allow_generate_tasks'] and self.params['is_generate_tasks']:
            generated_tasks = self._generate_tasks(generated_data_units)
            result['generated_tasks_count'] = len(generated_tasks)
        else:
            self.run.log_message('Generating tasks process has passed.')

        # Generate ground truths for the uploaded data.
        # TODO: Need to add ground truths generation logic later.
        if self.config['options']['allow_generate_ground_truths'] and self.params['is_generate_ground_truths']:
            generated_ground_truths = self._generate_ground_truths()
            result['generated_ground_truths_count'] = len(generated_ground_truths)
        else:
            self.run.log_message('Generating ground truths process has passed.')

        self.run.end_log()
        return result

    def _analyze_collection(self) -> Dict:
        """Analyze Synapse Collection Specifications.

        Returns:
            Dict: The file specifications of the collection.
        """

        # Initialize progress
        self.run.set_progress(0, 1, category='analyze_collection')

        client = self.run.client
        collection_id = self.params['collection']
        collection = client.get_dataset(collection_id)

        # Finish progress
        self.run.set_progress(1, 1, category='analyze_collection')

        return collection['file_specifications']

    def _validate_organized_files(self, file_specification_template: Dict, organized_files: List) -> bool:
        """Validate organized files from Uploader."""
        validator = FileSpecificationValidator(file_specification_template, organized_files)
        return validator.validate()

    def _upload_files(self, organized_files, organized_files_count: int) -> List:
        """Upload files to synapse-backend.

        Returns:
            Dict: The result of the upload.
        """
        # Initialize progress
        self.run.set_progress(0, organized_files_count, category='upload_data_files')
        self.run.log_message('Uploading data files...')

        client = self.run.client
        collection_id = self.params['collection']
        upload_result = []
        organized_files_count = len(organized_files)
        current_progress = 0
        for organized_file in organized_files:
            uploaded_data_file = client.upload_data_file(organized_file, collection_id)
            self.run.log_data_file(organized_file, UploadStatus.SUCCESS)
            upload_result.append(uploaded_data_file)
            self.run.set_progress(current_progress, organized_files_count, category='upload_data_files')
            current_progress += 1

        # Finish progress
        self.run.set_progress(organized_files_count, organized_files_count, category='upload_data_files')
        self.run.log_message('Upload data files completed.')

        return upload_result

    def _generate_data_units(self, uploaded_files: List, upload_result_count: int) -> List:
        """Generate data units for the uploaded data.

        TODO: make batch size configurable.

        Returns:
            Dict: The result of the generate data units process.
        """
        # Initialize progress
        self.run.set_progress(0, upload_result_count, category='generate_data_units')

        client = self.run.client

        generated_data_units = []
        current_progress = 0
        batches = get_batched_list(uploaded_files, 100)
        batches_count = len(batches)
        for batch in batches:
            created_data_units = client.create_data_units(batch)
            generated_data_units.append(created_data_units)
            self.run.set_progress(current_progress, batches_count, category='generate_data_units')
            current_progress += 1
            for created_data_unit in created_data_units:
                self.run.log_data_unit(created_data_unit['id'], UploadStatus.SUCCESS)

        # Finish progress
        self.run.set_progress(upload_result_count, upload_result_count, category='generate_data_units')

        return sum(generated_data_units, [])

    def _generate_tasks(self, generated_data_units: List) -> List:
        """Setup task with uploaded synapse-backend data units.

        TODO: make batch size configurable.
        """
        # Initialize progress
        self.run.set_progress(0, 1, category='generate_tasks')
        self.run.log_message('Generating tasks with data files...')

        # Prepare batches for processing
        client = self.run.client
        project_id = self.params['project']
        current_progress = 0

        # Generate tasks
        generated_tasks = []
        generated_data_units_count = len(generated_data_units)
        for data_unit in generated_data_units:
            tasks_data = []
            task_data = {'project': project_id, 'data_unit': data_unit['id']}
            tasks_data.append(task_data)
            if tasks_data:
                created_tasks = client.create_tasks(tasks_data)
                created_task_ids = [created_task['id'] for created_task in created_tasks]
                generated_tasks.append(created_task_ids)
                for created_task_id in created_task_ids:
                    self.run.log_task(created_task_id, UploadStatus.SUCCESS)

            self.run.set_progress(current_progress, generated_data_units_count, category='generate_tasks')
            current_progress += 1

        # Finish progress
        self.run.log_message('Generating tasks completed')
        self.run.set_progress(1, 1, category='generate_tasks')

        return sum(generated_tasks, [])

    def _generate_ground_truths(self):
        """Generate ground truths for the uploaded data.

        TODO: Need to add ground truths generation logic later.
        """
        # Initialize progress
        self.run.set_progress(0, 1, category='generate_ground_truths')
        self.run.log_message('Generating ground truths...')

        # Finish progress
        self.run.log_message('Generating ground truths completed')
        self.run.set_progress(1, 1, category='generate_ground_truths')



import json
from pathlib import Path

from synapse_sdk.i18n import gettext as _
from synapse_sdk.plugins.categories.registry import _REGISTERED_ACTIONS, register_actions
from synapse_sdk.plugins.enums import PluginCategory
from synapse_sdk.plugins.exceptions import ActionError
from synapse_sdk.utils.file import get_dict_from_file


def get_action(action, params_data, *args, **kwargs):
    if isinstance(params_data, str):
        try:
            params = json.loads(params_data)
        except json.JSONDecodeError:
            params = get_dict_from_file(params_data)
    else:
        params = params_data

    config_data = kwargs.pop('config', False)
    if config_data:
        if isinstance(config_data, str):
            config = read_plugin_config(plugin_path=config_data)
        else:
            config = config_data
    else:
        config = read_plugin_config()
    category = config['category']
    return get_action_class(category, action)(params, config, *args, **kwargs)


def get_action_class(category, action):
    register_actions()
    return _REGISTERED_ACTIONS[category][action]


def get_available_actions(category):
    register_actions()
    return list(_REGISTERED_ACTIONS[category].keys())


def get_plugin_categories():
    return [plugin_category.value for plugin_category in PluginCategory]


def read_plugin_config(plugin_path=None):
    config_file_name = 'config.yaml'
    if plugin_path:
        config_path = Path(plugin_path) / config_file_name
    else:
        config_path = config_file_name
    return get_dict_from_file(config_path)


def read_requirements(file_path):
    file_path = Path(file_path)
    if not file_path.exists():
        return None

    requirements = []
    for line in file_path.read_text().splitlines():
        stripped_line = line.strip()
        if stripped_line and not stripped_line.startswith('#'):
            requirements.append(stripped_line)
    return requirements


def run_plugin(
    action,
    params,
    plugin_config=None,
    plugin_path=None,
    modules=None,
    requirements=None,
    envs=None,
    debug=False,
    **kwargs,
):
    from synapse_sdk.plugins.models import PluginRelease

    if not envs:
        envs = {}

    if debug:
        if plugin_path and plugin_path.startswith('http'):
            if not plugin_config:
                raise ActionError({'config': _('"plugin_path"가 url인 경우에는 "config"가 필수입니다.')})
            plugin_release = PluginRelease(config=plugin_config)
        else:
            plugin_release = PluginRelease(plugin_path=plugin_path)
            plugin_config = plugin_release.config

        if action not in plugin_release.actions:
            raise ActionError({'action': _('해당 액션은 존재하지 않습니다.')})

        if plugin_path:
            envs['SYNAPSE_DEBUG_PLUGIN_PATH'] = plugin_path

        if modules:
            envs['SYNAPSE_DEBUG_MODULES'] = ','.join(modules)

    else:
        if plugin_config is None:
            raise ActionError({'config': _('플러그인 설정은 필수입니다.')})

    action = get_action(
        action,
        params,
        config=plugin_config,
        requirements=requirements,
        envs=envs,
        debug=debug,
        **kwargs,
    )
    return action.run_action()



from upath import UPath

from synapse_sdk.utils.storage.providers import BaseStorage


class GCPStorage(BaseStorage):
    def __init__(self, url):
        super().__init__(url)

        self.upath = UPath(f'gs://{self.query_params["bucket_name"]}', token=self.query_params['credentials'])

    def get_pathlib(self, path):
        return self.upath.joinuri(path)



from upath import UPath

from synapse_sdk.utils.storage.providers import BaseStorage


class SFTPStorage(BaseStorage):
    def get_pathlib(self, path):
        credentials = self.query_params['params']
        host = self.query_params['host']
        root_path = self.query_params['root_path']

        username = credentials['username']
        password = credentials['password']
        if path == '/':
            path = ''
        return UPath(f'sftp://{host}', username=username, password=password) / root_path / path

    def get_path_file_count(self, pathlib_obj: UPath):
        """Get file count in the path from SFTP provider.

        Args:
            pathlib_obj (UPath): The path to get file count.

        Returns:
            int: The file count in the path.
        """
        count = 0
        files = list(pathlib_obj.glob('**/*'))
        for file in files:
            if file.is_file():
                count += 1
        return count

    def get_path_total_size(self, pathlib_obj: UPath):
        """Get total size of the files in the path from SFTP provider.

        Args:
            pathlib_obj (UPath): The path to get total size.

        Returns:
            int: The total size of the files in the path.
        """
        total_size = 0
        for file in pathlib_obj.glob('**/*'):
            if file.is_file():
                total_size += file.stat().st_size
        return total_size



import pytest

from synapse_sdk.clients.validators.collections import FileSpecificationValidator


@pytest.fixture
def single_image_file_spec_template():
    return [
        {
            'name': 'image_1',
            'description': '',
            'file_type': 'image',
            'file_type_display': '이미지',
            'configuration': {},
            'is_sequential': False,
            'is_required': True,
            'is_primary': True,
            'function_type': 'main',
            'index': 1,
        }
    ]


@pytest.fixture
def single_image_organized_files():
    original_files = []
    for i in range(1, 21):
        file_name = f'{i:04d}.jpg'
        original_files.append({
            'files': {
                'image_1': '',
            },
            'meta': {
                'original_file_name': file_name,
            },
        })
    return original_files


@pytest.fixture
def image_and_meta_file_spec_template():
    return [
        {
            'name': 'image_1',
            'description': '',
            'file_type': 'image',
            'file_type_display': '이미지',
            'configuration': {},
            'is_sequential': False,
            'is_required': True,
            'is_primary': True,
            'function_type': 'main',
            'index': 1,
        },
        {
            'name': 'image_2',
            'description': '',
            'file_type': 'image',
            'file_type_display': '이미지',
            'configuration': {},
            'is_sequential': False,
            'is_required': True,
            'is_primary': True,
            'function_type': 'main',
            'index': 2,
        },
        {
            'name': 'meta_1',
            'description': '',
            'file_type': 'text',
            'file_type_display': '텍스트',
            'configuration': {},
            'is_sequential': False,
            'is_required': True,
            'is_primary': False,
            'function_type': 'meta',
            'index': 1,
        },
    ]


@pytest.fixture
def image_and_meta_organized_files():
    original_files = []
    for i in range(1, 21):  # Create 20 copies
        image_1_file_name = f'{i:04d}_1.jpg'
        image_2_file_name = f'{i:04d}_2.jpg'
        meta_file_name = f'{i:04d}.txt'
        original_files.append({
            'files': {
                'image_1': f'/path/to/{image_1_file_name}',
                'image_2': f'/path/to/{image_2_file_name}',
                'meta_1': f'/path/to/{meta_file_name}',
            },
            'meta': {
                'original_file_name': {
                    'image_1': image_1_file_name,
                    'image_2': image_2_file_name,
                    'meta_1': meta_file_name,
                },
            },
        })
    return original_files


@pytest.fixture
def invalid_image_and_meta_organized_files():
    original_files = []
    for i in range(1, 21):  # Create 20 copies
        image_1_file_name = f'{i:04d}_1.jpg'
        image_2_file_name = f'{i:04d}_2.jpg'
        meta_file_name = f'{i:04d}.txt'
        original_files.append({
            'files': {
                'image_invalid_1': f'/path/to/{image_1_file_name}',
                'image_invalid_2': f'/path/to/{image_2_file_name}',
                'meta_1': f'/path/to/{meta_file_name}',
            },
            'meta': {
                'original_file_name': {
                    'image_1': image_1_file_name,
                    'image_2': image_2_file_name,
                    'meta_1': meta_file_name,
                },
            },
        })
    return original_files


def test_validate_single_image_file_spec_template(single_image_file_spec_template, single_image_organized_files):
    validator = FileSpecificationValidator(single_image_file_spec_template, single_image_organized_files)

    assert validator.validate() is True


def test_validate_image_and_meta_file_spec_template_success(
    image_and_meta_file_spec_template, image_and_meta_organized_files
):
    validator = FileSpecificationValidator(image_and_meta_file_spec_template, image_and_meta_organized_files)

    assert validator.validate() is True


def test_validate_image_and_meta_file_spec_template_failed_with_invalid_organized_files(
    image_and_meta_file_spec_template, invalid_image_and_meta_organized_files
):
    validator = FileSpecificationValidator(image_and_meta_file_spec_template, invalid_image_and_meta_organized_files)

    assert validator.validate() is False



import click

from .utils import load_alias, set_default_alias


@click.command()
@click.argument('name', required=True)
def default(name):
    """Set an alias as the default."""
    alias_data = load_alias(name)
    if not alias_data:
        click.echo(f"Alias '{name}' does not exist.")
        return

    set_default_alias(name)
    click.echo(f"Alias '{name}' is now the default.")



from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod
from synapse_sdk.plugins.exceptions import ActionError
from synapse_sdk.plugins.utils import get_action


@register_action
class AutoLabelAction(Action):
    name = 'auto_label'
    category = PluginCategory.SMART_TOOL
    method = RunMethod.TASK

    def get_auto_label(self):
        return self.entrypoint(**self.params)

    def run_model(self, input_data):
        try:
            action = get_action(
                'inference',
                input_data,
                config={
                    'category': 'neural_net',
                    'code': self.params['plugin'],
                    'version': self.params['version'],
                    'actions': {'inference': {'method': input_data['method']}},
                },
            )
            return action.run_action()
        except ActionError as e:
            raise Exception(e.errors)

    def start(self):
        auto_label = self.get_auto_label()
        input_data = auto_label.handle_input(self.params)
        output_data = self.run_model(input_data)
        return auto_label.handle_output(output_data)



import copy
import tempfile
from decimal import Decimal
from pathlib import Path
from typing import Annotated

from pydantic import AfterValidator, BaseModel, field_validator
from pydantic_core import PydanticCustomError

from synapse_sdk.clients.exceptions import ClientError
from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod
from synapse_sdk.plugins.models import Run
from synapse_sdk.utils.file import archive, get_temp_path, unarchive
from synapse_sdk.utils.pydantic.validators import non_blank


class TrainRun(Run):
    def log_metric(self, category, key, value, **metrics):
        # TODO validate input via plugin config
        self.log('metric', {'category': category, 'key': key, 'value': value, 'metrics': metrics})

    def log_visualization(self, category, group, image, **meta):
        # TODO validate input via plugin config
        self.log('visualization', {'category': category, 'group': group, **meta}, file=image)


class Hyperparameter(BaseModel):
    batch_size: int
    epochs: int
    learning_rate: Decimal


class TrainParams(BaseModel):
    name: Annotated[str, AfterValidator(non_blank)]
    description: str
    checkpoint: int | None
    dataset: int
    hyperparameter: Hyperparameter

    @field_validator('name')
    @staticmethod
    def unique_name(value, info):
        action = info.context['action']
        client = action.client
        try:
            model_exists = client.exists('list_models', params={'name': value})
            job_exists = client.exists(
                'list_jobs',
                params={
                    'ids_ex': action.job_id,
                    'category': 'neural_net',
                    'job__action': 'train',
                    'is_active': True,
                    'params': f'name:{value}',
                },
            )
            assert not model_exists and not job_exists, '존재하는 학습 이름입니다.'
        except ClientError:
            raise PydanticCustomError('client_error', '')
        return value


@register_action
class TrainAction(Action):
    name = 'train'
    category = PluginCategory.NEURAL_NET
    method = RunMethod.JOB
    run_class = TrainRun
    params_model = TrainParams
    progress_categories = {
        'dataset': {
            'proportion': 20,
        },
        'train': {
            'proportion': 75,
        },
        'model_upload': {
            'proportion': 5,
        },
    }

    def start(self):
        hyperparameter = self.params['hyperparameter']

        # download dataset
        self.run.log_message('Preparing dataset for training.')
        input_dataset = self.get_dataset()

        # retrieve checkpoint
        checkpoint = None
        if self.params['checkpoint']:
            self.run.log_message('Retrieving checkpoint.')
            checkpoint = self.get_model(self.params['checkpoint'])

        # train dataset
        self.run.log_message('Starting model training.')
        result = self.entrypoint(self.run, input_dataset, hyperparameter, checkpoint=checkpoint)

        # upload model_data
        self.run.log_message('Registering model data.')
        self.run.set_progress(0, 1, category='model_upload')
        model = self.create_model(result)
        self.run.set_progress(1, 1, category='model_upload')

        self.run.end_log()
        return {'model_id': model['id'] if model else None}

    def get_dataset(self):
        client = self.run.client
        assert bool(client)

        input_dataset = {}

        ground_truths, count_dataset = client.list_ground_truth_events(
            params={
                'fields': ['category', 'files', 'data'],
                'expand': ['data'],
                'ground_truth_dataset_versions': self.params['dataset'],
            },
            list_all=True,
        )
        self.run.set_progress(0, count_dataset, category='dataset')
        for i, ground_truth in enumerate(ground_truths, start=1):
            self.run.set_progress(i, count_dataset, category='dataset')
            try:
                input_dataset[ground_truth['category']].append(ground_truth)
            except KeyError:
                input_dataset[ground_truth['category']] = [ground_truth]

        return input_dataset

    def get_model(self, model_id):
        model = self.client.get_model(model_id)
        model_file = Path(model['file'])
        output_path = get_temp_path(f'models/{model_file.stem}')
        if not output_path.exists():
            unarchive(model_file, output_path)
        model['path'] = output_path
        return model

    def create_model(self, path):
        params = copy.deepcopy(self.params)
        configuration_fields = ['hyperparameter']
        configuration = {field: params.pop(field) for field in configuration_fields}

        with tempfile.TemporaryDirectory() as temp_path:
            input_path = Path(path)
            archive_path = Path(temp_path, 'archive.zip')
            archive(input_path, archive_path)

            return self.client.create_model({
                'plugin': self.plugin_release.plugin,
                'version': self.plugin_release.version,
                'file': str(archive_path),
                'configuration': configuration,
                **params,
            })



from synapse_sdk.clients.base import BaseClient
from synapse_sdk.clients.utils import get_default_url_conversion


class HITLClientMixin(BaseClient):
    def get_assignment(self, pk):
        path = f'assignments/{pk}/'
        return self._get(path)

    def list_assignments(self, params=None, url_conversion=None, list_all=False):
        path = 'assignments/'
        url_conversion = get_default_url_conversion(url_conversion, files_fields=['files'])
        return self._list(path, params=params, url_conversion=url_conversion, list_all=list_all)

    def set_tags_assignments(self, data, params=None):
        path = 'assignments/set_tags/'
        return self._post(path, payload=data, params=params)



import tempfile

import jwt
from fastapi import FastAPI
from ray import serve

from synapse_sdk.clients.backend import BackendClient
from synapse_sdk.utils.file import unarchive

app = FastAPI()


class BaseInference:
    backend_url = None
    client = None

    def __init__(self, backend_url):
        self.backend_url = backend_url

    @serve.multiplexed()
    async def _load_model(self, model_id: str):
        model_info = jwt.decode(model_id, self.backend_url, algorithms='HS256')
        client = BackendClient(self.backend_url, token=model_info['token'], tenant=model_info['tenant'])
        model = client.get_model(model_info['model'])
        with tempfile.TemporaryDirectory() as temp_path:
            unarchive(model['file'], temp_path)
            model['path'] = temp_path
            return await self._get_model(model)

    async def get_model(self):
        return await self._load_model(serve.get_multiplexed_model_id())

    async def _get_model(self, model):
        raise NotImplementedError

    async def infer(self, *args, **kwargs):
        raise NotImplementedError



from synapse_sdk.clients.base import BaseClient
from synapse_sdk.clients.exceptions import ClientError


class CoreClientMixin(BaseClient):
    def list_nodes(self, params=None):
        path = 'nodes'
        response = self._get(path, params=params)
        if not response['result']:
            raise ClientError(200, response['msg'])
        return response['data']['summary']

    def get_node(self, pk, params=None):
        path = f'nodes/{pk}'
        response = self._get(path, params=params)['detail']

        if not response['result']:
            raise ClientError(200, response['msg'])

        if 'agent' not in response['data']:
            raise ClientError(404, 'Node Not Found')
        return response['data']



def get_default_url_conversion(url_conversion, **kwargs):
    defaults = {'files_fields': [], 'coerce': None, 'is_list': True}
    defaults.update(kwargs)
    if url_conversion:
        defaults.update(url_conversion)
    return defaults


def get_batched_list(object_list, batch_size):
    return [object_list[index : index + batch_size] for index in range(0, len(object_list), batch_size)]



from synapse_sdk.clients.backend.annotation import AnnotationClientMixin
from synapse_sdk.clients.backend.core import CoreClientMixin
from synapse_sdk.clients.backend.dataset import DatasetClientMixin
from synapse_sdk.clients.backend.hitl import HITLClientMixin
from synapse_sdk.clients.backend.integration import IntegrationClientMixin
from synapse_sdk.clients.backend.ml import MLClientMixin


class BackendClient(
    AnnotationClientMixin,
    CoreClientMixin,
    DatasetClientMixin,
    IntegrationClientMixin,
    MLClientMixin,
    HITLClientMixin,
):
    name = 'Backend'
    token = None
    tenant = None
    agent_token = None

    def __init__(self, base_url, token=None, tenant=None, agent_token=None):
        super().__init__(base_url)
        self.token = token
        self.tenant = tenant
        self.agent_token = agent_token

    def _get_headers(self):
        headers = {}
        if self.token:
            headers['Authorization'] = f'Token {self.token}'
        if self.tenant:
            headers['SYNAPSE-Tenant'] = f'Token {self.tenant}'
        if self.agent_token:
            headers['SYNAPSE-Agent'] = f'Token {self.agent_token}'
        return headers



class FileSpecificationValidator:
    """File specification validator class for synapse backend collection.

    Args:
        file_spec_template (list):
            * List of dictionaries containing file specification template
            * This is from synapse-backend file specification data.
        organized_files (list): List of dictionaries containing organized files.
    """

    def __init__(self, file_spec_template, organized_files):
        self.file_spec_template = file_spec_template
        self.organized_files = organized_files

    def validate(self):
        """Validate the file specification template with organized files.

        Returns:
            bool: True if the file specification template is valid, False otherwise.
        """
        for spec in self.file_spec_template:
            spec_name = spec['name']
            is_required = spec['is_required']

            for file_group in self.organized_files:
                files = file_group['files']
                if is_required and spec_name not in files:
                    return False
                if spec_name in files and not files[spec_name]:
                    return False
        return True



from urllib.parse import parse_qs, urlparse


class BaseStorage:
    url = None
    options = None
    OPTION_CASTS = {}

    def __init__(self, connection_params: str | dict):
        self.url = None

        if isinstance(connection_params, dict):
            self.query_params = connection_params['configuration']
        else:
            self.url = urlparse(connection_params)
            self.query_params = self.url_querystring_to_dict()

    def url_querystring_to_dict(self):
        query_string = self.url.query

        query_dict = parse_qs(query_string)

        for key, value in query_dict.items():
            if len(value) == 1:
                query_dict[key] = value[0]

        return {
            key: self.OPTION_CASTS[key](value) if key in self.OPTION_CASTS else value
            for key, value in query_dict.items()
        }

    def upload(self, source, target):
        raise NotImplementedError

    def exists(self, target):
        raise NotImplementedError

    def get_url(self, target):
        raise NotImplementedError

    def get_pathlib(self, path):
        """Get the path as a pathlib object.

        Args:
            path (str): The path to convert.

        Returns:
            pathlib.Path: The converted path.
        """
        raise NotImplementedError

    def get_path_file_count(self, pathlib_obj):
        """Get the file count in the path.

        Args:
            pathlib_obj (UPath): The path to get file count.

        Returns:
            int: The file count in the path.
        """
        raise NotImplementedError

    def get_path_total_size(self, pathlib_obj):
        """Get the total size of the path.

        Args:
            pathlib_obj (UPath): The path to get total size.

        Returns:
            int: The total size of the path.
        """
        raise NotImplementedError



import pytest
from pydantic import ValidationError

from synapse_sdk.clients.backend.models import Storage
from synapse_sdk.clients.base import BaseClient


@pytest.fixture
def base_client():
    return BaseClient('http://fake_url')


@pytest.fixture
def valid_storage_response():
    return {
        'id': 1,
        'name': 'test_storage',
        'category': 'internal',
        'provider': 'file_system',
        'configuration': {},
        'is_default': True,
    }


@pytest.fixture
def invalid_storage_response(valid_storage_response):
    response = valid_storage_response.copy()
    response['provider'] = 'invalid_provider'
    return response


def test_validate_response_with_pydantic_model_success(base_client, valid_storage_response):
    validated_response = base_client._validate_response_with_pydantic_model(valid_storage_response, Storage)
    assert validated_response['id'] == valid_storage_response['id']


def test_validate_response_with_pydantic_model_invalid_data(base_client, invalid_storage_response):
    with pytest.raises(ValidationError) as exc_info:
        base_client._validate_response_with_pydantic_model(invalid_storage_response, Storage)
    assert '1 validation error' in str(exc_info.value)


def test_validate_response_with_pydantic_model_not_pydantic_model(base_client, valid_storage_response):
    with pytest.raises(TypeError) as exc_info:
        base_client._validate_response_with_pydantic_model(valid_storage_response, {})
    assert 'The provided model is not a pydantic model' in str(exc_info.value)



from synapse_sdk.clients.base import BaseClient


class CoreClientMixin(BaseClient):
    def health_check(self):
        path = 'health/'
        return self._get(path)

    def get_metrics(self, panel):
        path = f'metrics/{panel}/'
        return self._get(path)



import click

from .utils import load_alias, save_alias_field


@click.command()
@click.argument('name', required=True)
@click.argument('field', required=True)
@click.argument('value', required=True)
def update(name, field, value):
    """Update a specific field in an alias."""
    alias_data = load_alias(name)
    if not alias_data:
        click.echo(f"Alias '{name}' does not exist.")
        return
    save_alias_field(name, field.upper(), value)
    click.echo(f"Alias '{name}' updated. Field '{field}' is now '{value}'.")



import hashlib


def hash_text(text):
    md5_hash = hashlib.md5()
    md5_hash.update(text.encode('utf-8'))
    return md5_hash.hexdigest()


def str_to_bool(value):
    return value.lower() in ['true', '1', 'yes']



import gettext as _gettext
import locale
from pathlib import Path

ASSIGNED_LOCALE = None


def set_language(language):
    ASSIGNED_LOCALE = language  # noqa: F841


def get_locale():
    if ASSIGNED_LOCALE is not None:
        return ASSIGNED_LOCALE

    system_locale, encoding = locale.getlocale()

    if system_locale is None:
        system_locale = 'en'

    return system_locale


language = get_locale().split('_')[0]

PACKAGE_DIR = Path(__file__).parent
LOCALE_DIR = PACKAGE_DIR.parent / 'locale'

translation = _gettext.translation('messages', localedir=LOCALE_DIR, languages=[language], fallback=True)
translation.install()

gettext = translation.gettext
ngettext = translation.ngettext
pgettext = translation.pgettext
npgettext = translation.npgettext



from multiprocessing import Pool
from pathlib import Path
from typing import Dict, Optional

from tqdm import tqdm

from synapse_sdk.clients.base import BaseClient
from synapse_sdk.clients.utils import get_batched_list


class DatasetClientMixin(BaseClient):
    def list_dataset(self):
        path = 'datasets/'
        return self._list(path)

    def get_dataset(self, dataset_id):
        """Get dataset from synapse-backend.

        Args:
            dataset_id: The dataset id to get.
        """
        path = f'datasets/{dataset_id}/?expand=file_specifications'
        return self._get(path)

    def create_data_file(self, file_path: Path):
        """Create data file to synapse-backend.

        Args:
            file_path: The file pathlib object to upload.
        """
        path = 'data_files/'
        return self._post(path, files={'file': file_path})

    def create_data_units(self, data):
        """Create data units to synapse-backend.

        Args:
            data: The data bindings to upload from create_data_file interface.
        """
        path = 'data_units/'
        return self._post(path, data=data)

    def upload_dataset(
        self,
        dataset_id: int,
        dataset: Dict,
        project_id: Optional[int] = None,
        batch_size: int = 1000,
        process_pool: int = 10,
    ):
        """Upload dataset to synapse-backend.

        Args:
            dataset_id: The dataset id to upload the data to.
            dataset: The dataset to upload.
                * structure:
                    - files: The files to upload. (key: file name, value: file pathlib object)
                    - meta: The meta data to upload.
            project_id: The project id to upload the data to.
            batch_size: The batch size to upload the data.
            process_pool: The process pool to upload the data.
        """
        # TODO validate dataset with schema

        params = [(data, dataset_id) for data in dataset]

        with Pool(processes=process_pool) as pool:
            dataset = pool.starmap(self.upload_data_file, tqdm(params))

        batches = get_batched_list(dataset, batch_size)

        for batch in tqdm(batches):
            data_units = self.create_data_units(batch)

            if project_id:
                tasks_data = []
                for data, data_unit in zip(batch, data_units):
                    task_data = {'project': project_id, 'data_unit': data_unit['id']}
                    # TODO: Additional logic needed here if task data storage is required during import.

                    tasks_data.append(task_data)

                self.create_tasks(tasks_data)

    def upload_data_file(self, data: Dict, dataset_id: int) -> Dict:
        """Upload files to synapse-backend.

        Args:
            data: The data to upload.
                * structure:
                    - files: The files to upload. (key: file name, value: file pathlib object)
                    - meta: The meta data to upload.
            dataset_id: The dataset id to upload the data to.

        Returns:
            Dict: The result of the upload.
        """
        for name, path in data['files'].items():
            data_file = self.create_data_file(path)
            data['dataset'] = dataset_id
            data['files'][name] = {'checksum': data_file['checksum'], 'path': str(path)}
        return data



import os
from pathlib import Path

import click

from synapse_sdk.clients.agent import AgentClient
from synapse_sdk.clients.backend import BackendClient
from synapse_sdk.plugins.models import PluginRelease
from synapse_sdk.plugins.utils import read_requirements, run_plugin


@click.command()
@click.argument('action')
@click.argument('params')
@click.option('--job-id')
@click.option('--direct/--no-direct', default=False)
@click.option('--run-by', type=click.Choice(['script', 'agent', 'backend']), default='script')
@click.option('--agent-host')
@click.option('--agent-token')
@click.option('--host')
@click.option('--agent')
@click.option('--user-token')
@click.option('--tenant')
@click.pass_context
def run(ctx, action, params, job_id, direct, run_by, agent_host, agent_token, host, agent, user_token, tenant):
    debug = ctx.obj['DEBUG']

    if run_by == 'script':
        run_by_script(action, params, job_id, direct, debug)
    elif run_by == 'agent':
        run_by_agent(action, params, job_id, agent_host, agent_token, user_token, tenant, debug)
    elif run_by == 'backend':
        run_by_backend(action, params, agent, host, user_token, tenant)


def run_by_script(action, params, job_id, direct, debug):
    requirements = read_requirements('./requirements.txt')
    result = run_plugin(action, params, requirements=requirements, debug=debug, job_id=job_id, direct=direct)

    if debug:
        click.echo(result)


def run_by_agent(action, params, job_id, agent_host, agent_token, user_token, tenant, debug):
    client = AgentClient(agent_host, agent_token, user_token, tenant)
    data = {'action': action, 'params': params}
    plugin_path = os.getcwd()
    requirements = read_requirements(Path(plugin_path) / 'requirements.txt')
    if requirements:
        data = {'requirements': requirements}

    if job_id:
        data['job_id'] = job_id

    if debug:
        data.update({
            'plugin_path': plugin_path,
            'modules': os.getenv('SYNAPSE_DEBUG_MODULES', '').split(','),
        })
        result = client.run_debug_plugin_release(data=data)
    else:
        plugin_release = PluginRelease()
        result = client.run_plugin_release(plugin_release.code, data=data)

    click.echo(result)


def run_by_backend(action, params, agent, host, user_token, tenant):
    client = BackendClient(host, user_token, tenant=tenant)
    plugin_release = PluginRelease()
    data = {'agent': agent, 'version': plugin_release.version, 'action': action, 'params': params}
    result = client.run_plugin(plugin_release.plugin, data=data)

    click.echo(result)



import asyncio
import base64
import hashlib
import json
import mimetypes
import operator
import re
import zipfile
from functools import reduce
from pathlib import Path

import aiohttp
import requests
import yaml

from synapse_sdk.utils.network import clean_url
from synapse_sdk.utils.string import hash_text


def download_file(url, path_download, name=None, coerce=None, use_cached=True):
    chunk_size = 1024 * 1024 * 50
    cleaned_url = clean_url(url)  # remove query params and fragment

    if name:
        use_cached = False
    else:
        name = hash_text(cleaned_url)

    name += Path(cleaned_url).suffix

    path = Path(path_download) / name

    if not use_cached or not path.is_file():
        response = requests.get(url, allow_redirects=True, stream=True)
        response.raise_for_status()

        with path.open('wb') as file:
            for chunk in response.iter_content(chunk_size=chunk_size):
                file.write(chunk)

    if coerce:
        path = coerce(path)

    return path


def files_url_to_path(files, coerce=None, file_field=None):
    path_download = get_temp_path('media')
    path_download.mkdir(parents=True, exist_ok=True)
    if file_field:
        files[file_field] = download_file(files[file_field], path_download, coerce=coerce)
    else:
        for file_name in files:
            if isinstance(files[file_name], str):
                files[file_name] = download_file(files[file_name], path_download, coerce=coerce)
            else:
                files[file_name]['path'] = download_file(files[file_name].pop('url'), path_download, coerce=coerce)


def files_url_to_path_from_objs(objs, files_fields, coerce=None, is_list=False, is_async=False):
    if is_async:
        asyncio.run(afiles_url_to_path_from_objs(objs, files_fields, coerce=coerce, is_list=is_list))
    else:
        if not is_list:
            objs = [objs]

        for obj in objs:
            for files_field in files_fields:
                try:
                    files = reduce(operator.getitem, files_field.split('.'), obj)
                    if isinstance(files, str):
                        files_url_to_path(obj, coerce=coerce, file_field=files_field)
                    else:
                        files_url_to_path(files, coerce=coerce)
                except KeyError:
                    pass


async def adownload_file(url, path_download, name=None, coerce=None, use_cached=True):
    chunk_size = 1024 * 1024 * 50
    cleaned_url = clean_url(url)  # remove query params and fragment

    if name:
        use_cached = False
    else:
        name = hash_text(cleaned_url)

    name += Path(cleaned_url).suffix

    path = Path(path_download) / name

    if not use_cached or not path.is_file():
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                with path.open('wb') as file:
                    while chunk := await response.content.read(chunk_size):
                        file.write(chunk)

    if coerce:
        path = coerce(path)

    return path


async def afiles_url_to_path(files, coerce=None):
    path_download = get_temp_path('media')
    path_download.mkdir(parents=True, exist_ok=True)
    for file_name in files:
        if isinstance(files[file_name], str):
            files[file_name] = await adownload_file(files[file_name], path_download, coerce=coerce)
        else:
            files[file_name]['path'] = await adownload_file(files[file_name].pop('url'), path_download, coerce=coerce)


async def afiles_url_to_path_from_objs(objs, files_fields, coerce=None, is_list=False):
    if not is_list:
        objs = [objs]

    tasks = []

    for obj in objs:
        for files_field in files_fields:
            try:
                files = reduce(operator.getitem, files_field.split('.'), obj)
                tasks.append(afiles_url_to_path(files, coerce=coerce))
            except KeyError:
                pass

    await asyncio.gather(*tasks)


def get_dict_from_file(file_path):
    if isinstance(file_path, str):
        file_path = Path(file_path)

    with open(file_path) as f:
        if file_path.suffix == '.yaml':
            return yaml.safe_load(f)
        else:
            return json.load(f)


def calculate_checksum(file_path, prefix=''):
    md5_hash = hashlib.md5()
    with open(file_path, 'rb') as f:
        for byte_block in iter(lambda: f.read(4096), b''):
            md5_hash.update(byte_block)
    checksum = md5_hash.hexdigest()
    if prefix:
        return f'dev-{checksum}'
    return checksum


def archive(input_path, output_path, append=False):
    input_path = Path(input_path)
    output_path = Path(output_path)

    mode = 'a' if append and output_path.exists() else 'w'
    with zipfile.ZipFile(output_path, mode=mode, compression=zipfile.ZIP_DEFLATED) as zipf:
        if input_path.is_file():
            zipf.write(input_path, input_path.name)
        else:
            for file_path in input_path.rglob('*'):
                if file_path.is_file():  # Only add files, skip directories
                    arcname = file_path.relative_to(input_path.parent)
                    zipf.write(file_path, arcname)


def unarchive(file_path, output_path):
    """
    Unarchives a ZIP file to a given directory.

    Parameters:
        file_path (str | Path): The path to the ZIP file.
        output_path (str): The directory where the files will be extracted.
    """
    output_path = Path(output_path)
    output_path.mkdir(parents=True, exist_ok=True)

    with zipfile.ZipFile(str(file_path), 'r') as zip_ref:
        zip_ref.extractall(output_path)


def get_temp_path(sub_path=None):
    path = Path('/tmp/datamaker')
    if sub_path:
        path = path / sub_path
    return path


def convert_file_to_base64(file_path):
    """
    Convert a file to base64 using pathlib.

    Args:
        file_path (str): Path to the file to convert

    Returns:
        str: Base64 encoded string of the file contents
    """
    # FIXME base64 is sent sometimes.
    if file_path.startswith('data:'):
        return file_path

    # Convert string path to Path object
    path = Path(file_path)

    try:
        # Read binary content of the file
        binary_content = path.read_bytes()

        # Convert to base64
        base64_encoded = base64.b64encode(binary_content).decode('utf-8')

        # Get the MIME type of the file
        mime_type, _ = mimetypes.guess_type(path)
        assert mime_type is not None, 'MIME type cannot be guessed'

        # Convert bytes to string for readable output
        return f'data:{mime_type};base64,{base64_encoded}'

    except FileNotFoundError:
        raise FileNotFoundError(f'File not found: {file_path}')
    except Exception as e:
        raise Exception(f'Error converting file to base64: {str(e)}')



import contextlib
import subprocess
from functools import cached_property
from pathlib import Path

from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod


@register_action
class GradioAction(Action):
    name = 'gradio'
    category = PluginCategory.NEURAL_NET
    method = RunMethod.JOB

    @property
    def working_directory(self):
        dir = Path.cwd() / self.config['directory'].replace('.', '/')
        assert dir.is_dir(), f'Working directory {dir} does not exist.'
        return dir

    @property
    def _requirements(self):
        return self.config.get('requirements', ['gradio>=5'])

    @property
    def tag(self):
        _tag = f'{self.plugin_release.code}-{self.plugin_release.checksum}'
        return _tag.replace('@', '-')

    @cached_property
    def deploy_port(self):
        return self._get_avail_ports_host()

    def deploy(self):
        self.run.log('deploy', 'Start deploying')

        try:
            # Write Dockerfile and requirements.txt
            path_dockerfile = self.write_dockerfile_template()
            self.write_requirements(path_dockerfile.parent / 'requirements.txt')

            # Build docker image
            self.build_docker_image(path_dockerfile)

            # Run docker image
            self.run_docker_image()
        except Exception as e:
            self.run.log('deploy', f'Error: {e}')
            raise e

    def start(self):
        self.deploy()
        return {'endpoint': f'http://localhost:{self.deploy_port}'}

    def write_dockerfile_template(self):
        dockerfile_path = self.working_directory / 'Dockerfile'

        with open(dockerfile_path, 'w') as f:
            f.write("""FROM python:3.10
WORKDIR /home/user/app

RUN pip install --no-cache-dir pip -U && \\
    pip install --no-cache-dir uvicorn

RUN apt-get update && \\
    apt-get install -y curl && \\
    curl -fsSL https://deb.nodesource.com/setup_22.x | bash - && \\
    apt-get install -y nodejs && \\
    rm -rf /var/lib/apt/lists/* && \\
    apt-get clean

COPY . /home/user/app

RUN pip install --no-cache-dir -r requirements.txt

EXPOSE 7860

CMD ["python", "app.py"]
""")
        return dockerfile_path

    def write_requirements(self, path):
        with open(path, 'w') as f:
            f.write('\n'.join(self._requirements))

    def build_docker_image(self, path_dockerfile):
        self.run.log('deploy', 'Start building docker image')
        result = subprocess.run(
            [
                'docker',
                'build',
                '-t',
                self.tag,
                '-f',
                str(path_dockerfile),
                '.',
            ],
            cwd=self.working_directory,
            check=True,
        )
        print(result)

    @staticmethod
    def _get_avail_ports_host(start_port=8900, end_port=8999):
        import nmap

        nm = nmap.PortScanner()

        scan_range = f'{start_port}-{end_port}'
        nm.scan(hosts='host.docker.internal', arguments=f'-p {scan_range}')

        try:
            open_ports = nm['host.docker.internal']['tcp'].keys()
            open_ports = [int(port) for port in open_ports]
        except KeyError:
            open_ports = []

        for port in range(start_port, end_port + 1):
            if port not in open_ports:
                return port

        raise IOError(f'No free ports available in range {start_port}-{end_port}')

    def run_docker_image(self):
        self.run.log('deploy', 'Start running docker image')

        # Check for existing container
        self.run.log('deploy', 'Check for existing container')
        with contextlib.suppress(subprocess.CalledProcessError):
            subprocess.run(['docker', 'stop', self.tag], check=True)
            subprocess.run(['docker', 'rm', self.tag], check=True)

        # Run docker image
        self.run.log('deploy', 'Starting docker container')
        subprocess.run(
            [
                'docker',
                'run',
                '-d',
                '--name',
                self.tag,
                '-p',
                f'{self.deploy_port}:7860',
                self.tag,
            ],
            check=True,
        )



class ActionError(Exception):
    errors = None

    def __init__(self, errors, *args):
        if isinstance(errors, (str, dict)):
            self.errors = errors
        elif isinstance(errors, Exception) and len(errors.args) == 1:
            self.errors = ActionError(errors.args[0]).errors
        else:
            self.errors = str(errors)
        super().__init__(errors, *args)

    def as_drf_error(self, data=None):
        if data is None:
            data = self.errors

        if isinstance(data, list):
            return data
        elif isinstance(data, dict):
            return {key: self.as_drf_error(value) for key, value in data.items()}

        return [str(data)]



import click

from .dataclass import ENV_VARS
from .utils import CONFIG_DIR, DEFAULT_FILE, ensure_config_dir, save_alias_field, set_default_alias


@click.command()
def create():
    """Create a new alias."""
    ensure_config_dir()
    name = click.prompt('Enter alias name (eg. prod, test)', type=str)
    alias_file = CONFIG_DIR / f'{name}.env'
    if alias_file.exists():
        click.echo(f"Alias '{name}' already exists.")
        return
    if alias_file.name == DEFAULT_FILE.name:
        click.echo(f"Invalid alias name '{name}'.")
        return

    alias_data = {}

    for key, value in ENV_VARS.items():
        prompt_kwargs = {'text': value.name, 'type': value.type, 'show_default': value.show_default}
        if bool(value.default):
            prompt_kwargs['default'] = alias_data[value.default]

        alias_data[key] = click.prompt(**prompt_kwargs)

    for key, value in alias_data.items():
        save_alias_field(name, key, value)

    # Set the default alias if it is the first alias created
    if not DEFAULT_FILE.exists():
        set_default_alias(name)

    click.echo(f"Alias '{name}' created.")



import pytest
from pydantic import ValidationError

from synapse_sdk.clients.backend.models import Storage, StorageCategory, StorageProvider


def test_storage_model_validation_success():
    storage_response = {
        'id': 1,
        'name': 'test_storage',
        'category': 'internal',
        'provider': 'file_system',
        'configuration': {},
        'is_default': True,
    }
    storage = Storage(**storage_response)
    assert storage.id == 1


def test_storage_model_validation_failed_with_invalid_category():
    storage_response = {
        'id': 1,
        'name': 'test_storage',
        'category': 'invalid_data',
        'provider': 'file_system',
        'configuration': {},
        'is_default': True,
    }
    with pytest.raises(ValidationError) as exc_info:
        Storage(**storage_response)
    assert 'category' in str(exc_info.value)


def test_storage_model_validation_failed_with_invalid_provider():
    storage_response = {
        'id': 1,
        'name': 'test_storage',
        'category': 'internal',
        'provider': 'invalid_provider',
        'configuration': {},
        'is_default': True,
    }
    with pytest.raises(ValidationError) as exc_info:
        Storage(**storage_response)
    assert 'provider' in str(exc_info.value)


def test_storage_model_validation_failed_with_missing_field():
    storage_response = {
        'id': 1,
        'name': 'test_storage',
        'category': 'internal',
        'provider': 'file_system',
        'is_default': True,
        # Missing configuration field
    }
    with pytest.raises(ValidationError):
        Storage(**storage_response)


def test_storage_model_enum_values():
    # Test that valid enum values work correctly
    for category in StorageCategory:
        for provider in StorageProvider:
            storage = Storage(
                id=1,
                name='test_storage',
                category=category,
                provider=provider,
                configuration={},
                is_default=True,
            )
            assert storage.category == category
            assert storage.provider == provider



from synapse_sdk.clients.agent.core import CoreClientMixin
from synapse_sdk.clients.agent.ray import RayClientMixin
from synapse_sdk.clients.agent.service import ServiceClientMixin
from synapse_sdk.clients.exceptions import ClientError


class AgentClient(CoreClientMixin, RayClientMixin, ServiceClientMixin):
    name = 'Agent'
    agent_token = None
    user_token = None
    tenant = None
    long_poll_handler = None

    def __init__(self, base_url, agent_token, user_token=None, tenant=None, long_poll_handler=None):
        super().__init__(base_url)
        self.agent_token = agent_token
        self.user_token = user_token
        self.tenant = tenant
        self.long_poll_handler = long_poll_handler

    def _get_headers(self):
        headers = {'Authorization': self.agent_token}
        if self.user_token:
            headers['SYNAPSE-User'] = f'Token {self.user_token}'
        if self.tenant:
            headers['SYNAPSE-Tenant'] = f'Token {self.tenant}'
        return headers

    def _request(self, method, path, **kwargs):
        if self.long_poll_handler:
            return self._request_long_poll(method, path, **kwargs)
        return super()._request(method, path, **kwargs)

    def _request_long_poll(self, method, path, **kwargs):
        headers = self._get_headers()

        if kwargs.get('files'):
            raise ClientError(400, 'file is not allowed when long polling')

        headers['Content-Type'] = 'application/json'

        request_id = self.long_poll_handler.set_request({'method': method, 'path': path, 'headers': headers, **kwargs})
        try:
            response = self.long_poll_handler.get_response(request_id)
        except TimeoutError:
            raise ClientError(408, f'{self.name} is not responding')

        if 400 <= response['status'] < 600:
            raise ClientError(response['status'], response.json() if response['status'] == 400 else response['reason'])

        return response['data']



from pathlib import Path

from dotenv import dotenv_values, load_dotenv, set_key, unset_key

CONFIG_DIR = Path.home() / '.config' / 'synapse_alias'
DEFAULT_FILE = CONFIG_DIR / '__default__'


def ensure_config_dir():
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)


def load_alias(alias_name):
    ensure_config_dir()
    alias_file = CONFIG_DIR / alias_name
    if not alias_file.exists():
        return None
    return dotenv_values(alias_file)


# Save or update a field in an alias .env file
def save_alias_field(alias_name, key, value):
    ensure_config_dir()
    alias_file = CONFIG_DIR / alias_name
    set_key(alias_file, key, value)


# Delete a field from an alias .env file
def delete_alias_field(alias_name, key):
    ensure_config_dir()
    alias_file = CONFIG_DIR / alias_name
    unset_key(alias_file, key)


# Delete an alias .env file
def delete_alias_file(alias_name):
    ensure_config_dir()
    alias_file = CONFIG_DIR / alias_name
    if alias_file.exists():
        alias_file.unlink()


def get_default_alias():
    ensure_config_dir()
    if DEFAULT_FILE.exists():
        default_data = dotenv_values(DEFAULT_FILE)
        return default_data.get('DEFAULT_ALIAS')
    return None


# Set the default alias
def set_default_alias(alias_name):
    ensure_config_dir()
    with open(DEFAULT_FILE, 'w') as f:
        f.write(f'DEFAULT_ALIAS={alias_name}\n')


def load_dotenv_default_alias():
    default_alias = get_default_alias()
    if default_alias is not None:
        load_dotenv(CONFIG_DIR / default_alias)



from synapse_sdk.utils.storage.providers.gcp import GCPStorage
from synapse_sdk.utils.storage.providers.s3 import S3Storage
from synapse_sdk.utils.storage.providers.sftp import SFTPStorage

STORAGE_PROVIDERS = {
    's3': S3Storage,
    'amazon_s3': S3Storage,
    'minio': S3Storage,
    'gcp': GCPStorage,
    'sftp': SFTPStorage,
}



import logging


def remap_class_labels(dataset, data_type, class_pattern, new_class):
    """
    Remaps class labels in a dataset by replacing specific classification patterns.

    This function finds items of a specified data type (e.g., 'pcd', 'image', 'video')
    where the classification matches a given pattern (e.g., 'truck', 'car'), and
    remaps them to a new class label (e.g., 'vehicle'). Useful for consolidating
    or standardizing class labels in ground truth data.

    Args:
        dataset (dict): The dataset containing ground truth labels
        data_type (str): Type of data to process (e.g., 'pcd', 'image', 'video')
        class_pattern (dict): Pattern to identify target classifications (e.g., {'name': 'truck'})
        new_class (dict): New classification to apply (e.g., {'name': 'vehicle'})

    Returns:
        dict: Dataset with remapped class labels
    """
    updated_dataset = dataset.copy()

    if data_type not in updated_dataset:
        logging.log(logging.WARNING, f"Data type '{data_type}' not found in dataset.")
        return updated_dataset

    def matches_pattern(classification, pattern):
        """Check if a classification matches the specified pattern."""
        for key, pattern_value in pattern.items():
            if key not in classification:
                return False

            if isinstance(pattern_value, dict) and isinstance(classification[key], dict):
                if not matches_pattern(classification[key], pattern_value):
                    return False
            elif classification[key] != pattern_value:
                return False

        return True

    for item in updated_dataset[data_type]:
        if 'classification' in item and matches_pattern(item['classification'], class_pattern):
            item['classification'] = new_class.copy()

    return updated_dataset



import jwt

from synapse_sdk.plugins.categories.base import Action
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.enums import PluginCategory, RunMethod


@register_action
class InferenceAction(Action):
    name = 'inference'
    category = PluginCategory.NEURAL_NET
    method = RunMethod.RESTAPI

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        headers = {'serve_multiplexed_model_id': self.get_serve_multiplexed_model_id()}
        if 'headers' in self.params:
            self.params['headers'].update(headers)
        else:
            self.params['headers'] = headers

    def get_serve_multiplexed_model_id(self):
        return jwt.encode(
            {
                'model': str(self.params.pop('model')),
                'token': self.envs['SYNAPSE_PLUGIN_RUN_USER_TOKEN'],
                'tenant': self.envs['SYNAPSE_PLUGIN_RUN_TENANT'],
            },
            self.envs['SYNAPSE_PLUGIN_RUN_HOST'],
        )



class MyAutoLabel:
    def __init__(self, **kwargs):
        pass

    def handle_input(self, input_data):
        """smart tool의 input을 model의 input 형태로 변환"""
        return

    def handle_output(self, output_data):
        """model의 output을 smart tool의 output 형태로 변환"""
        return



from typing import Any

from pydantic import HttpUrl
from pydantic_core import core_schema
from pydantic_core.core_schema import ValidationInfo

from synapse_sdk.utils.file import download_file, get_temp_path


class FileField(str):
    @classmethod
    def __get_pydantic_core_schema__(cls, source_type: Any, _handler: callable) -> core_schema.CoreSchema:
        return core_schema.with_info_before_validator_function(cls.validate, core_schema.str_schema())

    @staticmethod
    def validate(url: HttpUrl, info: ValidationInfo) -> str:
        path_download = get_temp_path('media')
        path_download.mkdir(parents=True, exist_ok=True)
        return str(download_file(url, path_download))



import shutil
from pathlib import Path

import yaml


def copy_project_category_template(category):
    copy_plugin(category)
    merge_config(category)


def copy_plugin(category):
    template_path = Path(__file__).parent / category / 'templates' / 'plugin'
    if not template_path.exists():
        return

    output_path = Path('plugin')
    shutil.copytree(template_path, output_path, dirs_exist_ok=True)


def merge_config(category):
    config_path = Path(__file__).parent / category / 'templates' / 'config.yaml'
    if not config_path.exists():
        return

    config_base_path = Path('config.yaml')

    config_base = yaml.safe_load(config_base_path.read_text(encoding='utf-8'))
    config = yaml.safe_load(config_path.read_text(encoding='utf-8'))

    config_base.update(config)
    config_base_path.write_text(yaml.dump(config_base, sort_keys=False), encoding='utf-8')



import click

from .utils import load_alias


@click.command()
@click.argument('name', required=True)
def read(name):
    """Read an alias."""
    alias_data = load_alias(name)
    if not alias_data:
        click.echo(f"Alias '{name}' does not exist.")
        return
    for key, value in alias_data.items():
        click.echo(f'{key}: {value}')



import copy
import tempfile
from pathlib import Path
from typing import Annotated, Optional

from pydantic import AfterValidator, BaseModel, field_validator
from pydantic_core import PydanticCustomError

from synapse_sdk.clients.exceptions import ClientError
from synapse_sdk.plugins.categories.decorators import register_action
from synapse_sdk.plugins.categories.neural_net.actions.train import TrainAction, TrainRun
from synapse_sdk.plugins.enums import PluginCategory, RunMethod
from synapse_sdk.utils.file import archive
from synapse_sdk.utils.module_loading import import_string
from synapse_sdk.utils.pydantic.validators import non_blank


class TuneRun(TrainRun):
    is_tune = True
    completed_samples = 0
    num_samples = 0
    checkpoint_output = None


class TuneConfig(BaseModel):
    mode: Optional[str] = None
    metric: Optional[str] = None
    num_samples: int = 1
    max_concurrent_trials: Optional[int] = None


class TuneParams(BaseModel):
    name: Annotated[str, AfterValidator(non_blank)]
    description: str
    checkpoint: int | None
    dataset: int
    tune_config: TuneConfig

    @field_validator('name')
    @staticmethod
    def unique_name(value, info):
        action = info.context['action']
        client = action.client
        try:
            job_exists = client.exists(
                'list_jobs',
                params={
                    'ids_ex': action.job_id,
                    'category': 'neural_net',
                    'job__action': 'tune',
                    'is_active': True,
                    'params': f'name:{value}',
                },
            )
            assert not job_exists, '존재하는 튜닝 작업 이름입니다.'
        except ClientError:
            raise PydanticCustomError('client_error', '')
        return value


@register_action
class TuneAction(TrainAction):
    """
    **Must read** Important notes before using Tune:

    1. Path to the model output (which is the return value of your train function)
       should be set to the checkpoint_output attribute of the run object **before**
       starting the training.
    2. Before exiting the training function, report the results to Tune.
    3. When using own tune.py, take note of the difference in the order of parameters.
       tune() function starts with hyperparameter, run, dataset, checkpoint, **kwargs
       whereas the train() function starts with run, dataset, hyperparameter, checkpoint, **kwargs.
    ----
    1)
    Set the output path for the checkpoint to export best model

    output_path = Path('path/to/your/weights')
    run.checkpoint_output = str(output_path)

    2)
    Before exiting the training function, report the results to Tune.
    The results_dict should contain the metrics you want to report.

    Example: (In train function)
    results_dict = {
        "accuracy": accuracy,
        "loss": loss,
        # Add other metrics as needed
    }
    if hasattr(self.dm_run, 'is_tune') and self.dm_run.is_tune:
        tune.report(results_dict, checkpoint=tune.Checkpoint.from_directory(self.dm_run.checkpoint_output))


    3)
    tune() function takes hyperparameter, run, dataset, checkpoint, **kwargs in that order
    whereas train() function takes run, dataset, hyperparameter, checkpoint, **kwargs in that order.

    --------------------------------------------------------------------------------------------------------

    **중요** Tune 사용 전 반드시 읽어야 할 사항들

    1. 본 플러그인의 train 함수에서, 학습을 진행하기 코드 전에
       결과 모델 파일의 경로(train함수의 리턴 값)을 checkpoint_output 속성에 설정해야 합니다.
    2. 학습이 종료되기 전에, 결과를 Tune에 보고해야 합니다.
    3. 플러그인에서 tune.py를 직접 생성해서 사용할 시, 매개변수의 순서가 다릅니다.

    ----
    1)
    체크포인트를 설정할 경로를 지정합니다.
    output_path = Path('path/to/your/weights')
    run.checkpoint_output = str(output_path)

    2)
    학습이 종료되기 전에, 결과를 Tune에 보고합니다.
    results_dict = {
        "accuracy": accuracy,
        "loss": loss,
        # 필요한 다른 메트릭 추가
    }
    if hasattr(self.dm_run, 'is_tune') and self.dm_run.is_tune:
        tune.report(results_dict, checkpoint=tune.Checkpoint.from_directory(self.dm_run.checkpoint_output))

    3)
    tune() 함수는 hyperparameter, run, dataset, checkpoint, **kwargs 순서이고
    train() 함수는 run, dataset, hyperparameter, checkpoint, **kwargs 순서입니다.
    """

    name = 'tune'
    category = PluginCategory.NEURAL_NET
    method = RunMethod.JOB
    run_class = TuneRun
    params_model = TuneParams
    progress_categories = {
        'dataset': {
            'proportion': 5,
        },
        'trials': {
            'proportion': 90,
        },
        'model_upload': {
            'proportion': 5,
        },
    }

    def start(self):
        from ray import tune

        # download dataset
        self.run.log_message('Preparing dataset for hyperparameter tuning.')
        input_dataset = self.get_dataset()

        # retrieve checkpoint
        checkpoint = None
        if self.params['checkpoint']:
            self.run.log_message('Retrieving checkpoint.')
            checkpoint = self.get_model(self.params['checkpoint'])

        # train dataset
        self.run.log_message('Starting training for hyperparameter tuning.')

        # Save num_samples to TuneRun for logging
        self.run.num_samples = self.params['tune_config']['num_samples']

        entrypoint = self.entrypoint
        if not self._tune_override_exists():
            # entrypoint must be train entrypoint
            def _tune(param_space, run, dataset, checkpoint=None, **kwargs):
                return entrypoint(run, dataset, param_space, checkpoint, **kwargs)

            entrypoint = _tune

        trainable = tune.with_parameters(entrypoint, run=self.run, dataset=input_dataset, checkpoint=checkpoint)
        tune_config = self.params['tune_config']

        hyperparameter = self.params['hyperparameter']
        param_space = self.convert_tune_params(hyperparameter)
        temp_path = tempfile.TemporaryDirectory()

        tuner = tune.Tuner(
            tune.with_resources(trainable, resources=self.tune_resources),
            tune_config=tune.TuneConfig(**tune_config),
            run_config=tune.RunConfig(
                name=f'synapse_tune_hpo_{self.job_id}',
                log_to_file=('stdout.log', 'stderr.log'),
                storage_path=temp_path.name,
            ),
            param_space=param_space,
        )
        result = tuner.fit()

        best_result = result.get_best_result()

        # upload model_data
        self.run.log_message('Registering best model data.')
        self.run.set_progress(0, 1, category='model_upload')
        self.create_model_from_result(best_result)
        self.run.set_progress(1, 1, category='model_upload')

        self.run.end_log()

        return {'best_result': best_result.config}

    @property
    def tune_resources(self):
        resources = {}
        for option in ['num_cpus', 'num_gpus']:
            option_value = self.params.get(option)
            if option_value:
                # Remove the 'num_' prefix and trailing s from the option name
                resources[(lambda s: s[4:-1])(option)] = option_value
        return resources

    def create_model_from_result(self, result):
        params = copy.deepcopy(self.params)
        configuration_fields = ['hyperparameter']
        configuration = {field: params.pop(field) for field in configuration_fields}

        with tempfile.TemporaryDirectory() as temp_path:
            archive_path = Path(temp_path, 'archive.zip')

            # Archive tune results
            # https://docs.ray.io/en/latest/tune/tutorials/tune_get_data_in_and_out.html#getting-data-out-of-tune-using-checkpoints-other-artifacts
            archive(result.path, archive_path)

            return self.client.create_model({
                'plugin': self.plugin_release.plugin,
                'version': self.plugin_release.version,
                'file': str(archive_path),
                'configuration': configuration,
                **params,
            })

    @staticmethod
    def convert_tune_params(param_list):
        """
        Convert YAML hyperparameter configuration to Ray Tune parameter dictionary.

        Args:
            param_list (list): List of hyperparameter configurations.

        Returns:
            dict: Ray Tune parameter dictionary
        """
        from ray import tune

        param_space = {}

        for param in param_list:
            name = param['name']
            param_type = param['type']

            if param_type == 'loguniform':
                param_space[name] = tune.loguniform(param['min'], param['max'])
            elif param_type == 'choice':
                param_space[name] = tune.choice(param['options'])
            elif param_type == 'randint':
                param_space[name] = tune.randint(param['min'], param['max'])
            # Add more type handlers as needed

        return param_space

    @staticmethod
    def _tune_override_exists(module_path='plugin.tune') -> bool:
        try:
            import_string(module_path)
            return True
        except ImportError:
            return False



import pkgutil
from importlib import import_module

from synapse_sdk.plugins.enums import PluginCategory

_REGISTERED_ACTIONS = {}


def register_actions():
    if not _REGISTERED_ACTIONS:
        for category in PluginCategory:
            plugin_category_module_name = f'synapse_sdk.plugins.categories.{category.value}.actions'
            plugin_category_module = import_module(plugin_category_module_name)
            for _, action_name, _ in pkgutil.iter_modules(plugin_category_module.__path__):
                action_module_name = f'{plugin_category_module_name}.{action_name}'
                import_module(action_module_name)



import json
import os
from pathlib import Path

import requests

from synapse_sdk.clients.exceptions import ClientError
from synapse_sdk.utils.file import files_url_to_path_from_objs


class BaseClient:
    name = None
    base_url = None
    page_size = 100

    def __init__(self, base_url):
        self.base_url = base_url
        requests_session = requests.Session()
        self.requests_session = requests_session

    def _get_url(self, path):
        if not path.startswith(self.base_url):
            return os.path.join(self.base_url, path)
        return path

    def _get_headers(self):
        return {}

    def _request(self, method: str, path: str, **kwargs) -> dict | str:
        """Request handler for all HTTP methods.

        Args:
            method (str): HTTP method to use.
            path (str): URL path to request.
            **kwargs: Additional keyword arguments to pass to the request.

        Returns:
            dict | str: JSON response or text response.
        """
        url = self._get_url(path)
        headers = self._get_headers()
        headers.update(kwargs.pop('headers', {}))

        # List to store opened files to close after request
        opened_files = []

        if method in ['post', 'put', 'patch']:
            # If files are included in the request, open them as binary files
            if kwargs.get('files') is not None:
                for name, file in kwargs['files'].items():
                    # Handle both string and Path object cases
                    if isinstance(file, str):
                        file = Path(file)
                    if isinstance(file, Path):
                        opened_file = file.open(mode='rb')
                        kwargs['files'][name] = (file.name, opened_file)
                        opened_files.append(opened_file)
                if 'data' in kwargs:
                    for name, value in kwargs['data'].items():
                        if isinstance(value, dict):
                            kwargs['data'][name] = json.dumps(value)
            else:
                headers['Content-Type'] = 'application/json'
                if 'data' in kwargs:
                    kwargs['data'] = json.dumps(kwargs['data'])

        try:
            # Send request
            response = getattr(self.requests_session, method)(url, headers=headers, **kwargs)
            if not response.ok:
                raise ClientError(
                    response.status_code, response.json() if response.status_code == 400 else response.reason
                )
        except requests.ConnectionError:
            raise ClientError(408, f'{self.name} is not responding')

        # Close all opened files
        for opened_file in opened_files:
            opened_file.close()

        return self._post_response(response)

    def _post_response(self, response):
        try:
            return response.json()
        except ValueError:
            return response.text

    def _get(self, path, url_conversion=None, pydantic_model=None, **kwargs):
        """
        Perform a GET request and optionally convert response to a pydantic model.

        Args:
            path (str): URL path to request.
            url_conversion (dict, optional): Configuration for URL to path conversion.
            pydantic_model (Type, optional): Pydantic model to convert the response to.
            **kwargs: Additional keyword arguments to pass to the request.

        Returns:
            The response data, optionally converted to a pydantic model.
        """
        response = self._request('get', path, **kwargs)

        if url_conversion:
            if url_conversion['is_list']:
                files_url_to_path_from_objs(response['results'], **url_conversion, is_async=True)
            else:
                files_url_to_path_from_objs(response, **url_conversion)

        if pydantic_model:
            return self._validate_response_with_pydantic_model(response, pydantic_model)

        return response

    def _post(self, path, pydantic_model=None, **kwargs):
        response = self._request('post', path, **kwargs)
        if pydantic_model:
            return self._validate_response_with_pydantic_model(response, pydantic_model)
        else:
            return response

    def _put(self, path, pydantic_model=None, **kwargs):
        response = self._request('put', path, **kwargs)
        if pydantic_model:
            return self._validate_response_with_pydantic_model(response, pydantic_model)
        else:
            return response

    def _patch(self, path, pydantic_model=None, **kwargs):
        response = self._request('patch', path, **kwargs)
        if pydantic_model:
            return self._validate_response_with_pydantic_model(response, pydantic_model)
        else:
            return response

    def _delete(self, path, pydantic_model=None, **kwargs):
        response = self._request('delete', path, **kwargs)
        if pydantic_model:
            return self._validate_response_with_pydantic_model(response, pydantic_model)
        else:
            return response

    def _list(self, path, url_conversion=None, list_all=False, **kwargs):
        response = self._get(path, **kwargs)
        if list_all:
            return self._list_all(path, url_conversion, **kwargs), response['count']
        else:
            return response

    def _list_all(self, path, url_conversion=None, params={}, **kwargs):
        params['page_size'] = self.page_size
        response = self._get(path, url_conversion, params=params, **kwargs)
        yield from response['results']
        if response['next']:
            yield from self._list_all(response['next'], url_conversion, **kwargs)

    def exists(self, api, *args, **kwargs):
        return getattr(self, api)(*args, **kwargs)['count'] > 0

    def _validate_response_with_pydantic_model(self, response, pydantic_model):
        """Validate a response with a pydantic model."""
        # Check if model is a pydantic model (has the __pydantic_model__ attribute)
        if (
            hasattr(pydantic_model, '__pydantic_model__')
            or hasattr(pydantic_model, 'model_validate')
            or hasattr(pydantic_model, 'parse_obj')
        ):
            pydantic_model.model_validate(response)
            return response
        else:
            # Not a pydantic model
            raise TypeError('The provided model is not a pydantic model')



import datetime
import time

from synapse_sdk.clients.exceptions import ClientError


class BaseLogger:
    progress_record = {}
    progress_categories = None
    current_category = None
    time_begin_per_category = {}

    def __init__(self, progress_categories=None):
        self.progress_categories = progress_categories
        if progress_categories:
            self.progress_record['categories'] = progress_categories

    def set_progress(self, current, total, category=None):
        assert 0 <= current <= total and total > 0
        assert category is not None or 'categories' not in self.progress_record

        percent = (current / total) * 100
        percent = round(percent, 2)
        # TODO current 0 으로 시작하지 않아도 작동되도록 수정
        if current == 0:
            self.time_begin_per_category[category] = time.time()
            time_remaining = None
        else:
            seconds_per_item = (time.time() - self.time_begin_per_category[category]) / current
            time_remaining = round(seconds_per_item * (total - current), 2)

        current_progress = {'percent': percent, 'time_remaining': time_remaining}

        if category:
            self.current_category = category
            self.progress_record['categories'][category].update(current_progress)
        else:
            self.progress_record.update(current_progress)

    def get_current_progress(self):
        categories = self.progress_record.get('categories')

        if categories:
            category_progress = None

            overall = 0
            for category, category_record in categories.items():
                if category == self.current_category:
                    break
                overall += category_record['proportion']

            category_record = categories[self.current_category]
            category_percent = category_record.get('percent', 0)
            if not category_progress and 'percent' in category_record:
                category_progress = {
                    'category': self.current_category,
                    'percent': category_percent,
                    'time_remaining': category_record.get('time_remaining'),
                }
            if category_percent > 0:
                overall += round(category_record['proportion'] / 100 * category_percent, 2)
            progress = {'overall': overall, **category_progress}
        else:
            progress = {
                'overall': self.progress_record.get('percent'),
                'time_remaining': self.progress_record.get('time_remaining'),
            }

        return progress

    def log(self, action, data, file=None):
        raise NotImplementedError


class ConsoleLogger(BaseLogger):
    def set_progress(self, current, total, category=None):
        super().set_progress(current, total, category=category)
        print(self.get_current_progress())

    def log(self, action, data, file=None):
        print(action, data)


class BackendLogger(BaseLogger):
    logs_queue = []
    client = None
    job_id = None

    def __init__(self, client, job_id, **kwargs):
        super().__init__(**kwargs)
        self.client = client
        self.job_id = job_id

    def set_progress(self, current, total, category=None):
        super().set_progress(current, total, category=category)
        try:
            progress_record = {
                'record': self.progress_record,
                'current_progress': self.get_current_progress(),
            }
            self.client.update_job(self.job_id, data={'progress_record': progress_record})
        except ClientError:
            pass

    def log(self, event, data, file=None):
        print(event, data)

        log = {
            'event': event,
            'data': data,
            'datetime': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f'),
            'job': self.job_id,
        }
        if file:
            log['file'] = file

        self.logs_queue.append(log)

        try:
            self.client.create_logs(self.logs_queue)
            self.logs_queue.clear()
        except ClientError as e:
            print(e)



from synapse_sdk.clients.backend.models import Storage
from synapse_sdk.clients.base import BaseClient
from synapse_sdk.utils.file import convert_file_to_base64


class IntegrationClientMixin(BaseClient):
    def health_check_agent(self, token):
        path = f'agents/{token}/connect/'
        return self._post(path)

    def get_plugin(self, pk):
        path = f'plugins/{pk}/'
        return self._get(path)

    def create_plugin(self, data):
        path = 'plugins/'
        return self._post(path, data=data)

    def update_plugin(self, pk, data):
        path = f'plugins/{pk}/'
        return self._put(path, data=data)

    def run_plugin(self, pk, data):
        path = f'plugins/{pk}/run/'
        return self._post(path, data=data)

    def get_plugin_release(self, pk, params=None):
        path = f'plugin_releases/{pk}/'
        return self._get(path, params=params)

    def create_plugin_release(self, data):
        path = 'plugin_releases/'
        files = {'file': data.pop('file')}
        return self._post(path, data=data, files=files)

    def get_job(self, pk, params=None):
        path = f'jobs/{pk}/'
        return self._get(path, params=params)

    def list_jobs(self, params=None):
        path = 'jobs/'
        return self._list(path, params=params)

    def update_job(self, pk, data):
        path = f'jobs/{pk}/'
        return self._patch(path, data=data)

    def list_job_console_logs(self, pk):
        path = f'jobs/{pk}/console_logs/'
        return self._get(path)

    def tail_job_console_logs(self, pk):
        path = f'jobs/{pk}/tail_console_logs/'

        url = self._get_url(path)
        headers = self._get_headers()

        response = self.requests_session.get(url, headers=headers, stream=True)
        for line in response.iter_lines(decode_unicode=True):
            if line:
                yield f'{line}\n'

    def create_logs(self, data):
        path = 'logs/'
        if not isinstance(data, list):
            data = [data]

        for item in data:
            if 'file' in item:
                item['file'] = convert_file_to_base64(item['file'])

        return self._post(path, data=data)

    def create_serve_application(self, data):
        path = 'serve_applications/'
        return self._post(path, data=data)

    def list_serve_applications(self, params=None, list_all=False):
        path = 'serve_applications/'
        return self._list(path, params=params, list_all=list_all)

    def get_storage(self, pk):
        """Get specific storage data from synapse backend."""
        path = f'storages/{pk}/'
        return self._get(path, pydantic_model=Storage)



import hashlib
import os
from pathlib import Path

from synapse_sdk.clients.base import BaseClient


class CoreClientMixin(BaseClient):
    def create_chunked_upload(self, file_path):
        def read_file_in_chunks(file_path, chunk_size=1024 * 1024 * 50):
            with open(file_path, 'rb') as file:
                while chunk := file.read(chunk_size):
                    yield chunk

        file_path = Path(file_path)
        size = os.path.getsize(file_path)
        hash_md5 = hashlib.md5()

        url = 'chunked_upload/'
        offset = 0
        for chunk in read_file_in_chunks(file_path):
            hash_md5.update(chunk)
            data = self._put(
                url,
                data={'filename': file_path.name},
                files={'file': chunk},
                headers={'Content-Range': f'bytes {offset}-{offset + len(chunk) - 1}/{size}'},
            )
            offset = data['offset']
            url = data['url']

        return self._post(url, data={'md5': hash_md5.hexdigest()})
